/*
 * Copyright (C) 1996,1997,1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.control;


import java.util.Vector;
import java.util.Enumeration;
import java.net.UnknownHostException;
import java.io.IOException;

import ryo.mobautil.*;
import NET.shudo.moba.lang.MobaThread;
import NET.shudo.moba.lang.PlaceAddress;
import NET.shudo.moba.place.MobaPlace;
import NET.shudo.moba.place.Command;
import NET.shudo.moba.place.PlaceEventNotifier;

import ryo.orb.*;
import ryo.orb.registry.*;


/**
 *  place 椹륳ȥ顣<BR>
 * MOBAIO ʤɤѤ롣
 */
public
class ControlServer
	implements MobaPlaceController, PlaceEventNotifier, Runnable {
  transient private final Command command;
  transient private boolean debug = false;

  private final Vector listeners = new Vector();


  public ControlServer() {
    this.command = Command.getInstance();
  }


  public void run() {
    this.debug = MobaPlace.debug();

    try {
      ryo.orb.registry.LocateRegistry.createRegistry(PlaceAddress.controllerPort()).bind("MobaPlace", ryo.orb.ORB.exportObject(this));
    } catch (ORBException e) {
      e.printStackTrace();
    }
  }


  /**
   * listener Ͽ
   */
  public MobaThreadInfo[] addListener(MobaPlaceListener l) {
    if (l == null)  return new MobaThreadInfo[0];	// should return null

    synchronized (listeners) {
      if (!listeners.contains(l)) {
	listeners.addElement(l);
	if (debug) {
	  System.err.println("ControlServer: addListener " + l);
	  System.err.flush();
	}
      }
    }

    Thread[] thrs = MobaPlace.threads();
    MobaThreadInfo[] tinfos = infoArrayByThreadArray(thrs);
    if (debug) {
      System.err.print("Threads already exists:");
      for (int i = 0; i < tinfos.length; i++) {
	System.err.print(" " + tinfos[i]);
      }
      System.err.println();
    }

    return tinfos;
  }

  /**
   * ϿƤ listener 
   */
  public void removeListener(MobaPlaceListener l) {
    synchronized (listeners) {
      listeners.removeElement(l);
    }
    if (debug) {
      System.err.println("ControlServer: removeListener " + l);
      System.err.flush();
    }
  }


  /**
   * åɤư롣
   */
  public MobaThreadInfo[] start(
		MobaThreadInfo[] tinfos, MobaPlaceListener by) {
    int i, index;
    Vector threads = null;

    if (tinfos == null || tinfos.length <= 0)  return null;

    String[] args = new String[tinfos.length + 1];
    args[0] = null;

    index = 1;
    for (i = 0; i < tinfos.length; i++) {
      MobaThreadInfo tinfo = tinfos[i];
      if (tinfo == null)  continue;
      args[index++] = tinfo.className();
    }
    if (debug) {
      System.err.print("ControlServer:");
      for (i = 0; i < args.length; i++)
	System.err.print(" " + args[i]);
      System.err.println();
    }

    threads = command.start(args);

    tinfos = infoArrayByThreadVector(threads);
    startedNotify(tinfos, by);

    return tinfos;
  }


  public void startedNotify(MobaThread thr) {
    MobaThreadInfo[] tinfos = { infoByMobaThread(thr) };
    startedNotify(tinfos, null);
  }

  public void startedNotify(Vector threads) {
    // б MobaThreadInfo[] 
    MobaThreadInfo[] tinfos = infoArrayByThreadVector(threads);
    startedNotify(tinfos, null);
  }

  private void startedNotify(MobaThreadInfo[] tinfos, MobaPlaceListener by) {
    if (debug) {
      System.err.print("startedNotify:");
      for (int i = 0; i < tinfos.length; i++) {
	MobaThreadInfo tinfo = tinfos[i];
	if (tinfo == null)  System.err.println(" null");
	else  System.err.print(" " + tinfo.id());
      }
      System.err.print(" to");
    }

    // 
    synchronized (listeners) {
      for (Enumeration e = listeners.elements(); e.hasMoreElements(); ) {
	MobaPlaceListener l = (MobaPlaceListener)e.nextElement();
	if (l != null && !l.equals(by)) {
	  if (debug)
	    System.err.print(" " + l);
	  try { l.started(tinfos); }
	  catch (RemoteException ex) { listeners.removeElement(l); }
	}
      }      
    }

    if (debug)
      System.err.println(".");
  }


  /**
   * åɤߤ롣
   */
  public void stop(MobaThreadInfo[] tinfos, MobaPlaceListener by) {
    int i, index;

    if (tinfos == null || tinfos.length <= 0)  return;

    String[] args = new String[tinfos.length + 1];
    args[0] = null;

    index = 1;
    for (i = 0; i < tinfos.length; i++) {
      MobaThreadInfo tinfo = tinfos[i];
      if (tinfo == null)  continue;
      args[index++] = String.valueOf(tinfo.id());
    }
    if (debug) {
      System.err.print("ControlServer:");
      for (i = 0; i < args.length; i++)
	System.err.print(" " + args[i]);
      System.err.println();
    }

    command.stop(args);

    stoppedNotify(tinfos, by);

    System.gc();
  }


  public void stoppedNotify(MobaThread thr) {
    MobaThreadInfo[] tinfos = { infoByMobaThread(thr) };
    stoppedNotify(tinfos, null);
  }

  public void stoppedNotify(Vector threads) {
    // б MobaThreadInfo[] 
    MobaThreadInfo[] tinfos = infoArrayByThreadVector(threads);
    stoppedNotify(tinfos, null);
  }

  private void stoppedNotify(MobaThreadInfo[] tinfos, MobaPlaceListener by) {
    if (debug) {
      System.err.print("stoppedNotify:");
      for (int i = 0; i < tinfos.length; i++) {
	MobaThreadInfo tinfo = tinfos[i];
	if (tinfo == null)  System.err.println(" null");
	else  System.err.print(" " + tinfo.id());
      }
      System.err.print(" to");
    }

    // 
    synchronized (listeners) {
      for (Enumeration e = listeners.elements(); e.hasMoreElements(); ) {
	MobaPlaceListener l = (MobaPlaceListener)e.nextElement();
	if (l != null && !l.equals(by)) {
	  if (debug)
	    System.err.print(" " + l);
	  try { l.stopped(tinfos); }
	  catch (RemoteException ex) { listeners.removeElement(l); }
	}
      }
    }

    if (debug)
      System.err.println(".");
  }


  /**
   * åɤư롣
   */
  public void move(String host, int port,
		MobaThreadInfo[] tinfos, MobaPlaceListener by)
	throws NoSuchMobaPlaceException,
		NoSuchThreadException, CannotMoveThreadException {
    int i, index;

    if (tinfos == null || tinfos.length <= 0)  return;
    if (host == null)
      throw new NoSuchMobaPlaceException(this);

    // adjust port number from controller (ex. 10010) to place (ex. 10000)
    port -= PlaceAddress.controllerPortDiff();

    String[] args = new String[3];
    args[0] = null;
    args[2] = host + ":" + port;

    for (i = 0; i < tinfos.length; i++) {
      MobaThreadInfo tinfo = tinfos[i];
      if (tinfo == null)  continue;
      args[1] = String.valueOf(tinfo.id());
      if (debug) {
	System.err.print("ControlServer:");
	for (i = 0; i < args.length; i++)
	  System.err.print(" " + args[i]);
	System.err.println();
      }

      try {
	command.move(args);
      }
      catch (UnknownHostException e) {
	throw new NoSuchMobaPlaceException(this);
      }
      catch (NET.shudo.moba.place.NoSuchThreadException e) {
/*
	if (i > 0) {
	  MobaThreadInfo[] succeed = new MobaThreadInfo[i];
	  System.arraycopy(tinfos, 0, succeed, 0, i);
	  goneNotify(host, port, tinfos);
	}
*/
	throw new NoSuchThreadException(i);
      }
      catch (NET.shudo.moba.place.CannotMoveThreadException e) {
/*
	if (i > 0) {
	  MobaThreadInfo[] succeed = new MobaThreadInfo[i];
	  System.arraycopy(tinfos, 0, succeed, 0, i);
	  goneNotify(host, port, tinfos);
	}
*/
	throw new CannotMoveThreadException(i);
      }
    }

/*
    goneNotify(host, port, tinfos);
*/
  }


  public void goneNotify(String host, int port, MobaThreadInfo[] tinfos) {
    port += PlaceAddress.controllerPortDiff();

    if (debug) {
      System.err.print("goneNotify(" + host + ":" + port + "):");
      for (int i = 0; i < tinfos.length; i++) {
	MobaThreadInfo tinfo = tinfos[i];
	if (tinfo == null)  System.err.println(" null");
	else  System.err.print(" " + tinfo.id());
      }
      System.err.print(" to");
    }

    // 
    synchronized (listeners) {
      for (Enumeration e = listeners.elements(); e.hasMoreElements(); ) {
	MobaPlaceListener l = (MobaPlaceListener)e.nextElement();
/*
	if (l != null && !l.equals(by)) {
*/
	  if (debug)
	    System.err.print(" " + l);
	  try { l.gone(host, port, tinfos); }
	  catch (RemoteException ex) { listeners.removeElement(l); }
/*
	}
*/
      }
    }

    if (debug)
      System.err.println(".");
  }

  public void goneNotify(String host, int port, MobaThread thr) {
    if (thr == null)  return;

    MobaThreadInfo[] tinfos = { infoByMobaThread(thr) };

    goneNotify(host, port, tinfos);
  }

  public void comeNotify(String host, int port, MobaThreadInfo[] tinfos) {
    // port is not useful

    if (debug) {
      System.err.print("comeNotify(" + host + ":" + port + "):");
      for (int i = 0; i < tinfos.length; i++) {
	MobaThreadInfo tinfo = tinfos[i];
	if (tinfo == null)  System.err.println(" null");
	else  System.err.print(" " + tinfo.id());
      }
      System.err.print(" to");
    }

    // 
    synchronized (listeners) {
      for (Enumeration e = listeners.elements(); e.hasMoreElements(); ) {
	MobaPlaceListener l = (MobaPlaceListener)e.nextElement();
	if (debug)
	  System.err.print(" " + l);
	try { l.come(host, port, tinfos); }
	catch (RemoteException ex) { listeners.removeElement(l); }
      }
    }

    if (debug)
      System.err.println(".");
  }

  public void comeNotify(String host, int port, MobaThread thr) {
    if (thr == null)  return;

    MobaThreadInfo[] tinfos = { infoByMobaThread(thr) };

    comeNotify(host, port, tinfos);
  }


  private MobaThreadInfo[] infoArrayByThreadVector(Vector threads) {
    MobaThreadInfo[] ret = null;
    if (threads == null)  return null;

    ret = new MobaThreadInfo[threads.size()];
    int index = 0;
    for (Enumeration e = threads.elements(); e.hasMoreElements(); ) {
      Object t = e.nextElement();
      if (t instanceof MobaThread) {
	MobaThread mt = (MobaThread)t;
	ret[index++] = infoByMobaThread(mt);
      }
    }
    return ret;
  }

  private MobaThreadInfo[] infoArrayByThreadArray(Thread[] threads) {
    MobaThreadInfo[] ret = null;
    if (threads == null)  return null;

    ret = new MobaThreadInfo[threads.length];
    int index = 0;
    for (int i = 0; i < threads.length; i++) {
      Thread t = threads[i];
      if (t instanceof MobaThread) {
	MobaThread mt = (MobaThread)t;
	ret[index++] = infoByMobaThread(mt);
      }
    }
    return ret;
  }

  private MobaThreadInfo infoByMobaThread(MobaThread thr) {
    return new MobaThreadInfo(thr.id(),
		thr.runnableInstance().getClass().getName(),
		thr.getName());
  }
}
