/*
 * Copyright (C) 1996,1997,1998,1999,2003
 *	Kazuyuki Shudo
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

#include <stdio.h>	/* for NULL */
#include <stdlib.h>	/* for malloc(), calloc() */
#include <string.h>	/* for memcpy() */
#include <netinet/in.h>	/* for hton[ls],ntoh[ls]() */

#include <jni.h>	/* for native methods */

#include "moba.h"


#define MIN(A,B) (((A)<(B))?(A):(B))


/*
 * Local Functions
 */
static jbyteArray getBuffer(JNIEnv *env, jsize size);
static void flushOutputStream(JNIEnv *env,
			jobject /*java.io.OutputStream*/ out);

static void show64(char *x);


/*
 * Global Variables
 */
static jbyteArray *bufTable = NULL;
static jobject lock;
static int lockExists = 0;


/*
 * $B%P%C%U%!(B (byte[]) $B$r<hF@!"EPO?$9$k!#(B
 */
#define LOCK_CLASSNAME "NET/shudo/moba/io/IOLock"
#define BUF_TBLSIZE 7
static jbyteArray getBuffer(JNIEnv *env, jsize size) {
  jbyteArray array;

  if (!lockExists) {	/* lockExists $B$K$h$k87L)$JGSB>@)8f$O4|BT$7$F$$$J$$(B */
    lock = (jobject)(*env)->FindClass(env, LOCK_CLASSNAME);
    lock = (*env)->NewGlobalRef(env, lock);
    lockExists = 1;
  }

  (*env)->MonitorEnter(env, lock);
  {
    unsigned int hash;
    if (!bufTable)
      bufTable = (jbyteArray *)calloc(sizeof(jbyteArray),BUF_TBLSIZE);
    hash = (((unsigned int)env) >> 3) % BUF_TBLSIZE;
	/* (JNIEnv *) $B$,(B thread $B$4$H$K0[$J$k$3$H$r4|BT$7$F(B hash $BCM$KMxMQ(B */
    array = bufTable[hash];
    if (array) {
      if (size > (*env)->GetArrayLength(env, array))
	array = (*env)->NewByteArray(env, size);
    }
    else {	/* buffer $B:n@.(B */
      array = (*env)->NewByteArray(env, size);
#ifdef DEBUG
      if (!array) {
	printf("getBuffer(): Cannot allocate an array of byte.\n");
	fflush(stdout);
      }
#endif
      array = (*env)->NewGlobalRef(env, (jobject)array);
	/* $B$3$N(B Global ref. $B$r2rJ|$7$F$$$J$$(B! */
      if (array)
	bufTable[hash] = array;
    }
  }
  (*env)->MonitorExit(env, lock);

  return array;
}


#undef OPTIMIZE_NOT_JNI
#define BUF_SIZE 131072
/*
 * $B%P%$%HNs$r(B (OutputStream) out $B$K=q$-=P$9!#(B
 *
 * returns: $B@.8y(B: 0, $B<:GT(B: -1
 */
extern int writeBytes(JNIEnv *env,
		jobject /*java.io.OutputStream*/ out, char *data, size_t len) {
  jbyteArray array;

#ifdef WRITE_VERBOSE
  printf("writeBytes() (len:%d) called.\n", len);
  fflush(stdout);
#endif

  array = getBuffer(env, BUF_SIZE);

  (*env)->MonitorEnter(env, (jobject)array);
  {
    jsize writeSize;
#ifdef OPTIMIZE_NOT_JNI
    HArrayOfByte *hary = DeRef(env, array);
    char *body = unhand(hary)->body;
#endif

    while (len > 0) {
      writeSize = MIN(len, BUF_SIZE);
#ifdef WRITE_VERBOSE
      printf("  writeSize: %d\n", writeSize);
      fflush(stdout);
#endif
#ifdef OPTIMIZE_NOT_JNI
      memcpy(body, data, writeSize);
#else
      (*env)->SetByteArrayRegion(env, array, 0, writeSize, (jbyte *)data);
		/* memory copy... */
#endif
      if ((writeArrayOfByte(env, out, array, 0, writeSize)) < 0)  return -1;
		/* returns $B@.8y(B: 0, $B<:GT(B: -1 */

      data += writeSize;
      len -= writeSize;
    }
  }
  (*env)->MonitorExit(env, (jobject)array);

  return 0;
}


/*
 * $B%P%$%HNs$r(B (InputStream) in $B$+$iFI$_9~$`!#(B
 * $B%G!<%?NN0h$rMQ0U$7$F$*$/I,MW$O$J$$!#(B
 */
extern size_t readBytes(JNIEnv *env,
		jobject /*java.io.InputStream*/ in, char *data, size_t len) {
  jbyteArray array;
  size_t readSum;

  array = getBuffer(env, BUF_SIZE);

  (*env)->MonitorEnter(env, (jobject)array);
  {
    jsize readSize;
    jbyte *readData;
#ifdef OPTIMIZE_NOT_JNI
    HArrayOfByte *hary = DeRef(env, array);
    char *body = unhand(hary)->body;
#endif

    readSum = 0;
    while (len > 0) {
      readSize = MIN(len, BUF_SIZE);
      if ((readArrayOfByte(env, in, array, 0, readSize)) < 0)  return -1;
#ifdef OPTIMIZE_NOT_JNI
      memcpy(data, body, (size_t)readSize);
#else
      readData = (*env)->GetByteArrayElements(env, array, NULL);
      memcpy(data, readData, (size_t)readSize);
      (*env)->ReleaseByteArrayElements(env, array, readData, JNI_ABORT);
#endif

      data += readSize;
      len -= readSize;
      readSum += readSize;
    }
  }
  (*env)->MonitorExit(env, (jobject)array);

  return readSum;
}


/*
 * ObjectWriter#writeObject(Object) $B$r8F$S=P$9!#(B
 */
extern int writeObject(JNIEnv *env,
		jobject /*NET.shudo.moba.io.ObjectWriter*/ out, jobject obj) {
  static int methodExists = 0;
  static jmethodID method;

  if (!methodExists) {
    jclass clazz = (*env)->FindClass(env, WRITEOBJ_CLASSNAME);
    (*env)->NewGlobalRef(env, clazz);
    method = (*env)->GetMethodID(env, clazz,
			WRITEOBJ_METHOD, WRITEOBJ_SIGNATURE);
    if ((*env)->ExceptionOccurred(env)) {
      /* not reached. */
      printf("method " WRITEOBJ_SIGNATURE " " WRITEOBJ_METHOD
		"\n\t\tis not found.\n");
      fflush(stdout);
      return;
    }
    else  methodExists = 1;
  }

  (*env)->CallVoidMethod(env, out, method, obj);

  if ((*env)->ExceptionOccurred(env)) {
#ifdef DEBUG
    printf("writeObject(): An exception occurred.\n");
    fflush(stdout);
#endif
    return -1;
  }

  return 0;
}


/*
 * ObjectReader#readObject() $B$r8F$S=P$9!#(B
 */
extern jobject readObject(JNIEnv *env,
		jobject /*NET.shudo.moba.io.ObjectReader*/ in) {
  static int methodExists = 0;
  static jmethodID method;

  jobject result = NULL;

  if (!methodExists) {
    jclass clazz = (*env)->FindClass(env, READOBJ_CLASSNAME);
    (*env)->NewGlobalRef(env, clazz);
    method = (*env)->GetMethodID(env, clazz,
			READOBJ_METHOD, READOBJ_SIGNATURE);
    if ((*env)->ExceptionOccurred(env)) {
      /* not reached. */
      printf("method " READOBJ_SIGNATURE " " READOBJ_METHOD
		"\n\t\tis not found.\n");
      fflush(stdout);
      return;
    }
    else  methodExists = 1;
  }

  result = (*env)->CallObjectMethod(env, in, method);

  if ((*env)->ExceptionOccurred(env)) {
#ifdef DEBUG
    printf("readObject(): An exception occurred.\n");
    (*env)->ExceptionDescribe(env);
    fflush(stdout);
#endif
    return result;
  }

  return result;
}


/*
 * 64bit $B@0?t$r=q$-9~$`!#(B
 */
extern int writeInt64(JNIEnv *env,
		jobject /*java.io.OutputStream*/ out, jlong *p) {
#ifdef BIG_ENDIAN
  return writeBytes(env, out, (char *)p, 8);
#else
  jlong v = hton64(*p);
  return writeBytes(env, out, (char *)&v, 8);
#endif
}


/*
 * 32bit $B@0?t$r=q$-9~$`!#(B
 */
extern int writeInt32(JNIEnv *env,
		jobject /*java.io.OutputStream*/ out, jint *p) {
#ifdef BIG_ENDIAN
  return writeBytes(env, out, (char *)p, 4);
#else
  jint v;
  v = (jint)htonl(*(unsigned long *)p);
  return writeBytes(env, out, (char *)&v, 4);
#endif
}


/*
 * 16bit $B@0?t$r=q$-9~$`!#(B
 */
extern int writeInt16(JNIEnv *env,
		jobject /*java_io_OutputStream*/ out, jshort *p) {
#ifdef BIG_ENDIAN
  return writeBytes(env, out, (char *)p, 2);
#else
  jshort v;
  v = (jshort)htons(*(unsigned short *)p);
  return writeBytes(env, out, (char *)&v, 2);
#endif
}


/*
 * char $B$r=q$-9~$`!#(B
 */
extern int writeChar(JNIEnv *env,
		jobject /*java.io.OutputStream*/ out, int c) {
  char ch = (char)c;
  return writeBytes(env, out, &ch, 1);
}


static void show64(char *x) {
  int i;

  printf("0x%02x", *x++);
  for (i = 1; i < 8; i++) {
    printf(", 0x%02x", *x++);
  }
  printf("\n");
  fflush(stdout);
}
/*
 * 64bit $B@0?t$rFI$_9~$`!#(B
 */
extern jlong readInt64(JNIEnv *env, jobject /*java.io.InputStream*/ in) {
  jlong v;
  if (readBytes(env, in, (char *)&v, 8) != 8)
    return -1;
#ifdef BIG_ENDIAN
  return v;
#else
  return ntoh64(v);
#endif
}


/*
 * 32bit $B@0?t$rFI$_9~$`!#(B
 */
extern jint readInt32(JNIEnv *env, jobject /*java.io.InputStream*/ in) {
  jint ret = 0;

  if (readBytes(env, in, (char *)&ret, 4) != 4)
    return -1;
#ifdef BIG_ENDIAN
  return ret;
#else
  return (jint)ntohl(*(unsigned long *)&ret);
#endif
}


/*
 * 16bit $B@0?t$rFI$_9~$`!#(B
 */
extern jshort readInt16(JNIEnv *env, jobject /*java_io_InputStream*/ in) {
  jshort ret = 0;

  if (readBytes(env, in, (char *)&ret, 2) != 2)
    return -1;
#ifdef BIG_ENDIAN
  return ret;
#else
  return (jshort)ntohs(*(unsigned short *)&ret);
#endif
}


/*
 * char $B$rFI$_9~$`!#(B
 */
extern jbyte readChar(JNIEnv *env, jobject /*java.io.InputStream*/ in) {
  char ret;

  if (readBytes(env, in, &ret, 1) != 1)
    return -1;
  return ret;
}


/*
 * $BJ8;zNsD9(B, $BJ8;zNs$r=q$-=P$9!#(B
 *
 * returns: $B@.8y(B: 0, $B<:GT(B: -1
 */
extern int writeString(JNIEnv *env,
		jobject /*java.io.OutputStream*/ out, char *str) {
  size_t len = strlen(str);
  jshort len16 = (jshort)len;

  /* $BJ8;zNsD9(B */
  if ((writeInt16(env, out, &len16)) < 0)  return -1;

  /* $BJ8;zNs(B */
  if ((writeBytes(env, out, str, len)) < 0)  return -1;

  return 0;
}


/*
 * $BJ8;zNsD9(B, $BJ8;zNs$rFI$_9~$`!#(B
 */
extern char *readStrdup(JNIEnv *env, jobject /*java.io.InputStream*/ in) {
  jshort len16;
  size_t len;
  char *ptr;

  /* $BJ8;zNsD9(B */
  len16 = readInt16(env, in);
  len = len16 | (size_t)0;

  if (!(ptr = (char *)malloc(len + 1))) {
    SignalError(0, JAVAPKG "OutOfMemoryError", 0);
    return NULL;
  }

  /* $BJ8;zNs(B */
  if (readBytes(env, in, ptr, len) != len)
    return NULL;
  ptr[len] = '\0';	/* terminate */

  return ptr;
}


#define WRITEBYTE_CLASSNAME	"java/io/OutputStream"
#define WRITEBYTE_SIGNATURE	"([BII)V"
#define WRITEBYTE_METHOD	"write"
#define READBYTE_CLASSNAME	"java/io/InputStream"
#define READBYTE_SIGNATURE	"([BII)I"
#define READBYTE_METHOD		"read"
#define FLUSH_CLASSNAME		"java/io/OutputStream"
#define FLUSH_SIGNATURE		"()V"
#define FLUSH_METHOD		"flush"


/*
 * OutputStream#write(byte[], int, int) $B$r8F$S=P$9!#(B
 */
jint writeArrayOfByte(JNIEnv *env, jobject /*java.io.OutputStream*/ out,
	jbyteArray data, jint off, jint len) {
  static int methodExists = 0;
#ifdef OPTIMIZE_NOT_JNI
  ExecEnv *ee = JNIEnv2EE(env);
  JHandle *hout = (JHandle *)DeRef(env, out);
  JHandle *hdata = (JHandle *)DeRef(env, data);
#endif
  static jmethodID method;
#ifdef WRITE_VERBOSE
  printf("writeArrayOfByte() called.\n");
  fflush(stdout);
#endif

  if (!methodExists) {
    jclass clazz = (*env)->FindClass(env, WRITEBYTE_CLASSNAME);
    (*env)->NewGlobalRef(env, clazz);
    method = (*env)->GetMethodID(env, clazz,
			WRITEBYTE_METHOD, WRITEBYTE_SIGNATURE);
    if ((*env)->ExceptionOccurred(env)) {
      /* not reached. */
      printf("method " WRITEBYTE_SIGNATURE " " WRITEBYTE_METHOD " is not found.\n");
      fflush(stdout);
      return -1;
    }
    methodExists = 1;
  }

#ifdef OPTIMIZE_NOT_JNI
  do_execute_java_method(ee, hout,
	0, 0, (struct methodblock *)method, 0,
	/* argument */ hdata, off, len, NULL);
#else
  (*env)->CallVoidMethod(env, out, method, data, off, len);
#endif

  if ((*env)->ExceptionOccurred(env)) {
    /* (*env)->ExceptionClear(env); */
#ifdef DEBUG
    printf("writeArrayOfByte(): An exception occurred.\n");
    printf("  clazz of exc: %s\n",
	cbName(EE()->exception.exc->methods->classdescriptor));
    fflush(stdout);
#endif
    return -1;
  }
  else
    return 0;
}


/*
 * InputStream#read(byte[]) $B$r8F$S=P$9!#(B
 *
 * returns: $BFI$a$?%P%$%H?t(B, $B<:GT$7$?$i(B -1
 */
int readArrayOfByte(JNIEnv *env, jobject /*java.io.InputStream*/ in,
	jbyteArray data, jint off, jint len) {
  static int methodExists = 0;
  static jmethodID method;
  jint ret;
  jint readLen;

  if (!methodExists) {
    jclass clazz = (*env)->FindClass(env, READBYTE_CLASSNAME);
    (*env)->NewGlobalRef(env, clazz);
    method = (*env)->GetMethodID(env, clazz,
			READBYTE_METHOD, READBYTE_SIGNATURE);
    if ((*env)->ExceptionOccurred(env)) {
      /* not reached. */
      printf("method " READBYTE_SIGNATURE " " READBYTE_METHOD " is not found.\n");
      fflush(stdout);
      return -1;
    }
    else  methodExists = 1;
  }

  readLen = 0;
  while (1) {
    ret = (*env)->CallIntMethod(env, in, method, data, off, len);

#ifdef WRITE_VERBOSE
    {
      jbyte *body = (*env)->GetByteArrayElements(env, data, NULL);
      jsize bodylen = (*env)->GetArrayLength(env, data);
      int i;
      printf("readArrayOfByte(): ((byte[%d])data, %d, %d) returns %d.\n",
		bodylen, off, len, ret);
      for (i = 0; i < ret; i++)  printf(" %02x", (int)body[off + i] & 0xff);
      (*env)->ReleaseByteArrayElements(env, data, body, JNI_ABORT);
      printf("\n");  fflush(stdout);
    }
#endif

    if ((*env)->ExceptionOccurred(env)) {
#ifdef DEBUG
      printf("readArrayOfByte(): An exception occurred.\n");
      fflush(stdout);
#endif
      return -1;
    }

    if (ret < 0)  break;	/* EOF */

    len -= ret;  off += ret;
    readLen += ret;
    if (len <= 0)  break;
  }

  return readLen;
}


/*
 * Local Functions
 */

/*
 * OutputStream#flush() $B$r8F$S=P$9!#(B
 */
static void flushOutputStream(JNIEnv *env, jobject /*java.io.OutputStream*/ out) {
  static int methodExists = 0;
  static jmethodID method;

  if (!methodExists) {
    jclass clazz = (*env)->FindClass(env, FLUSH_CLASSNAME);
    (*env)->NewGlobalRef(env, clazz);
    method = (*env)->GetMethodID(env, clazz, FLUSH_METHOD, FLUSH_SIGNATURE);
    if ((*env)->ExceptionOccurred(env)) {
      /* not reached. */
      printf("method " FLUSH_SIGNATURE " " FLUSH_METHOD " is not found.\n");
      fflush(stdout);
      return;
    }
    else  methodExists = 1;
  }

  (*env)->CallVoidMethod(env, out, method);
#ifdef DEBUG
  if ((*env)->ExceptionOccurred(env)) {
    printf("flushOutputStream(): An exception occurred.\n");
    fflush(stdout);
  }
#endif
}
