/*
 * Copyright (C) 1996,1997,1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/*
 * MOBA  native methods 
 */


#ifndef _MOBA_H_
#define _MOBA_H_


#include <sys/types.h>	/* for size_t */
#include <jni.h>
#include <native.h>

#include "mobanative.h"


/*
 * Definitions
 */

/*
 * Don't touch below stuffs
 */

/* endian */
/* GNU libc /usr/include/bytesex.h  */
#undef LITTLE_ENDIAN
#undef BIG_ENDIAN
#if defined(__i386__) || defined(__alpha__) \
	|| (defined(__mips__) && (defined(MIPSEL) || defined (__MIPSEL__)))
# define LITTLE_ENDIAN
#elif defined(__mc68000__) || defined (__sparc__) || defined (__PPC__) \
	|| (defined(__mips__) && (defined(MIPSEB) || defined (__MIPSEB__)))
# define BIG_ENDIAN
#else
# error architecture not supported by the Linux C library
#endif


#define MOBA_PACKAGE	"NET/shudo/moba/"
#define MOBA_REFLECT_PACKAGE	MOBA_PACKAGE "reflect/"
#define MOBA_IO_PACKAGE		MOBA_PACKAGE "io/"

#undef JDK_STYLE	/* JDK  Object Serialization Ȥݤ */

#ifdef JDK_STYLE
#define OBJECT_PACKAGE "java/io/"
#define WRITEOBJ_CLASSNAME	OBJECT_PACKAGE "ObjectOutputStream"
#define READOBJ_CLASSNAME	OBJECT_PACKAGE "ObjectInputStream"
#else
#define OBJECT_PACKAGE MOBA_IO_PACKAGE
#define WRITEOBJ_CLASSNAME	OBJECT_PACKAGE "ObjectWriter"
#define READOBJ_CLASSNAME	OBJECT_PACKAGE "ObjectReader"
#endif

/* for util.c */
#define WRITEOBJ_SIGNATURE "(Ljava/lang/Object;)V"
#define READOBJ_SIGNATURE "()Ljava/lang/Object;"

#define WRITEOBJ_METHOD "writeObject"
#define READOBJ_METHOD "readObject"

/* for serialize.c */
#define OWN_WRITEOBJ_SIGNATURE "(L" WRITEOBJ_CLASSNAME ";)V"
#define OWN_READOBJ_SIGNATURE "(L" READOBJ_CLASSNAME ";)V"

#define OWN_WRITEOBJ_METHOD "writeObject"
#define OWN_READOBJ_METHOD "readObject"


/* frame->constant_pool μ */
#define CP_CP		1
#define CP_METHOD	2	/* in case that frame->returnpc is 0 */


/* stack_item μ */
#define STACK_ITEM_HANDLE	2
#define STACK_ITEM_OBJECT	3
#define STACK_ITEM_ADDRESS	4
#define STACK_ITEM_IMM32	5
#define STACK_ITEM_IMM64	6


/* for meta*.c */
#if (JDK_VER < 110)	/* old fashion */
  #define HMetaClassToClassClass(/*(JHandle *)*/ handle) \
	((ClassClass *)unhand(unhand(handle)->thisClass))
#else
  #define HMetaClassToClassClass(/*(JHandle *)*/ handle) \
	((ClassClass *)(unhand(handle)->thisClass))
#endif /* JDK_VER < 110 */


/*
 * Global Functions
 */
/* signal_md.c */
extern int validateAddress(void *addr);

/* signature.c */
extern int isSigObject(char *sig);
extern int isSigArray(char *sig);

/* util.c */
extern ClassClass *cp2class(cp_item_type *cp);

#define LocalVarTblLen(method)	((method)->localvar_table_length)
extern char *LocalVarName(struct methodblock *method, int index);
extern char *LocalVarSig(struct methodblock *method, int index);
#define LocalVarSlot(method, i)		(((method)->localvar_table)[i].slot)
#define LocalVarStartPC(method, i)	(((method)->localvar_table)[i].pc0)
#define LocalVarLengthPC(method, i)	(((method)->localvar_table)[i].length)
extern int isLocalVarAlive(struct methodblock *, int, uint32_t pcoff);

extern unsigned char *nextPC(unsigned char *pc);
extern void dumpOpcodeLength(FILE *out);

extern int isPtrHandle(void *p);

extern void outPC(FILE *out, unsigned char *pc, struct methodblock *method);
extern void outMethod(FILE *out, struct methodblock *method);

extern struct methodblock *pc2method(unsigned char *pc);
extern uint32_t pc2offset(struct methodblock *method, unsigned char *pc);
extern unsigned char *offset2pc(struct methodblock *method, uint32_t offset);
extern int16_t sp2offset(JavaFrame *frame, stack_item *sp);
extern stack_item *offset2sp(JavaFrame *frame, int16_t offset);
extern uint16_t cp2offset(ClassClass *clazz, cp_item_type *cp_item);
extern cp_item_type *offset2cp(ClassClass *clazz, uint16_t offset);

extern void *callMethod(JNIEnv *, jobject, jmethodID, int isStatic, ...);
extern jclass classLoaderFindClass(JNIEnv *, char *name, jobject loader);

#ifdef BIG_ENDIAN
#define ntoh64(X) (X)
#define hton64(X) (X)
#else
extern jlong hton64(jlong x);
extern jlong ntoh64(jlong x);
#endif

/* io.c */
jint writeArrayOfByte(JNIEnv *env, jobject /*java.io.OutputStream*/ out,
	jbyteArray data, jint off, jint len);
	/* call OutputStream.write(byte[], int, int) */
int readArrayOfByte(JNIEnv *env, jobject /*java.io.InputStream*/ in,
	jbyteArray data, jint off, jint len);
	/* call InputStream.read(byte[], int, int) */

extern int writeBytes(JNIEnv *env, jobject /*java.io.OutputStream*/ out,
			char *data, size_t len);
extern size_t readBytes(JNIEnv *env, jobject /*java.io.InputStream*/ in,
			char *data, size_t len);
extern int writeObject(JNIEnv *env,
		jobject /*NET.shudo.moba.io.ObjectWriter*/ out, jobject obj);
extern jobject readObject(JNIEnv *env,
		jobject /*NET.shudo.moba.io.ObjectReader*/ in);

extern int writeInt64(JNIEnv *env, jobject /*java.io.OutputStream*/ out, jlong *p);
extern int writeInt32(JNIEnv *env, jobject /*java.io.OutputStream*/ out, jint *p);
extern int writeInt16(JNIEnv *env, jobject /*java.io.OutputStream*/ out, jshort *p);
extern int writeChar(JNIEnv *env, jobject /*java.io.OutputStream*/ out, int c);
extern jlong readInt64(JNIEnv *env, jobject /*java.io.InputStream*/ in);
extern jint readInt32(JNIEnv *env, jobject /*java.io.InputStream*/ in);
extern jshort readInt16(JNIEnv *env, jobject /*java.io.InputStream*/ in);
extern jbyte readChar(JNIEnv *env, jobject /*java.io.InputStream*/ in);

extern int writeString(JNIEnv *env, jobject /*java.io.OutputStream*/ out, char *);
extern char *readStrdup(JNIEnv *env, jobject /*java.io.InputStream*/ in);


#endif /* _MOBA_H_ */
