/*
 * Copyright (C) 1996,1997,1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.place;


import java.net.Socket;
import java.net.InetAddress;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.BufferedOutputStream;	// for optimization of throughput
import java.io.FileOutputStream;	// for debug
import java.io.IOException;
import NET.shudo.moba.lang.MobaThread;
import NET.shudo.moba.lang.PlaceAddress;
import NET.shudo.moba.io.ClassWriter;
import NET.shudo.moba.io.ThreadWriter;


/**
 * $B%9%l%C%I$rAw$j=P$9%/%i%9!#(B<BR>
 * $BD>@\$O(B MobaThread#goTo(), MobaThread#moveTo(),
 * MobaThread#fork(), MobaThread#forkThisThread() $B$r;H$&$Y$-!#(B
 *
 * @see NET.shudo.moba.lang.MobaThread#goTo
 * @see NET.shudo.moba.lang.MobaThread#fork
 */
public
class ThreadExporter {
  /**
   * $BM?$($i$l$?%9%l%C%I$rB>$N7W;;5!$K0\F0$5$;$k!#(B<BR>
   * $BD>@\$O!"(BMobaThread.moveTo(), goTo() $B$rMxMQ$9$Y$-!#(B
   *
   * @see NET.shudo.moba.lang.MobaThread#goTo()
   * @see NET.shudo.moba.lang.MobaThread#moveTo()
   */
  public static void moveTo(MobaThread thread, PlaceAddress addr)
	throws IOException {
    try {
      ThreadExporter.fork(thread, addr);
    }
    catch (IOException e) {
      // fork failed.
      // $B85%9%l%C%I$O;_$a$J$$(B
      throw e;
    }

    if (!thread.isInternalized()) {
      // notify
      PlaceEventNotifier notifier = MobaPlace.eventNotifier();
      if (notifier != null)
	notifier.goneNotify(addr.hostName(), addr.port(), thread);

      // $B0\F085$N%9%l%C%I$r;_$a$k(B
      try {
	thread.stop();
      }
      catch (SecurityException e) {
	// not reached.
	e.printStackTrace();  System.exit(1);
      }

      System.gc();
    }

    Logger.log(null, "sending a thread succeed");
  }


  /**
   * $BM?$($i$l$?%9%l%C%I$rB>$N7W;;5!$K%3%T!<$9$k!#(B(cf. fork(2))
   * $BD>@\$O!"(BMobaThread.moveTo(), goTo() $B$rMxMQ$9$Y$-!#(B
   *
   * @see NET.shudo.moba.lang.MobaThread#goTo()
   * @see NET.shudo.moba.lang.MobaThread#moveTo()
   *
   * @return true to the origianl thread,
   * false to the thread created on remote machine.
   */
  public static boolean fork(MobaThread thread, PlaceAddress addr)
	throws IOException {
    Thread crntThread = Thread.currentThread();

//System.out.println("fork(" + thread + ", " + addr + ") called.");

//try{
//(new ThreadWriter(new FileOutputStream("output"))).writeObject(thread);
//}
//catch (Exception e) { e.printStackTrace(); System.exit(1); }


    int threadPort = addr.port() + PlaceAddress.THREAD_PORT_DIFF;
    Socket sock = null;
    sock = new Socket(addr.inetAddress(), threadPort);	// throws IOException

//    ThreadWriter out = new ThreadWriter(
//	new BufferedOutputStream(sock.getOutputStream(), 1024 * 32));
    ThreadWriter out = new ThreadWriter(sock.getOutputStream());


    if (!thread.equals(crntThread)) {
//System.out.println("move the other thread.");
      try {
        thread.suspend();
      }
      catch (SecurityException e) {
	throw new IOException("Security Exception caused.");
      }
    }
    else {
//System.out.println("move myself.");
    }


    // $B%9%l%C%I$,@8@.$5$l$?>l=j$r$d$j$H$j(B
    {
      PlaceAddress genedPlace = thread.generatedPlace();
      String str = genedPlace.toString();
      byte[] bary = str.getBytes();
      int len = bary.length;
      out.write((len >>> 8)& 0xff);
      out.write(len & 0xff);
      out.write(bary);
    }
    out.flush();


    // $B%9%l%C%I$N=q$-=P$7(B
    out.writeObject(thread);		// throws IOException


    boolean ret = !thread.isInternalized();
    if (ret) {	// $B85$N%9%l%C%I$G$O??$K$J$k(B
      out.flush();
      sock.close();
    }

    return ret;
  }
}
