/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.reflect;

import java.lang.reflect.Method;
import java.lang.reflect.Constructor;


public
class MetaMethod {
  private Class thisClass;
  private String thisSig;
  private String thisName;
  private int thisAccess;


  protected MetaMethod(Class clazz, Method method) {
    this.thisClass = clazz;

    this.thisSig = signatureOfMethod(method);
    this.thisName = method.getName();
    this.thisAccess = method.getModifiers();

    this.methodID = methodID(clazz, thisName, thisSig);
  }

  protected MetaMethod(Class clazz, Constructor method) {
    this.thisClass = clazz;

    this.thisSig = signatureOfConstructor(method);
    this.thisName = "<init>";
    this.thisAccess = method.getModifiers();

    this.methodID = methodID(clazz, thisName, thisSig);
  }


  private String signatureOfMethod(Method method) {
    StringBuffer buf = new StringBuffer();

    Class retType = method.getReturnType();
    Class[] paramTypes = method.getParameterTypes();
    int nparam = paramTypes.length;

    buf.append('(');
    if (nparam > 0) {
      buf.append(MetaClass.signatureByClass(paramTypes[0]));
      for (int i = 1; i < nparam; i++) {
	buf.append(MetaClass.signatureByClass(paramTypes[i]));
      }
    }
    buf.append(')');

    buf.append(MetaClass.signatureByClass(retType));

    return buf.toString();
  }

  private String signatureOfConstructor(Constructor method) {
    StringBuffer buf = new StringBuffer();

    Class[] paramTypes = method.getParameterTypes();
    int nparam = paramTypes.length;

    buf.append('(');
    if (nparam > 0) {
      buf.append(MetaClass.signatureByClass(paramTypes[0]));
      for (int i = 1; i < nparam; i++) {
	buf.append(MetaClass.signatureByClass(paramTypes[i]));
      }
    }
    buf.append(")V");

    return buf.toString();
  }

  public String name() { return thisName; }
  public String signature() { return thisSig; }
  public int access() { return thisAccess; }

  private int methodID;		// JNI  jmethodID Ǽ
  public int hashCode() { return methodID; }
  public int ID() { return methodID; }
  protected static native int methodID(Class clazz, String name, String sig);

  public native void invoke(Object obj, Object[] args)
	throws IllegalArgumentException;
}
